@@ -7,7 +7,7 @@ OCR-D CLI: workspace management
 """
 import os
 from os import getcwd
-from os.path import relpath, exists, join, isabs, dirname, basename, abspath
+from os.path import relpath, exists, join, isabs
 from pathlib import Path
 from json import loads
 import sys
@@ -18,7 +18,7 @@ import time
 import click
 
 from ocrd import Resolver, Workspace, WorkspaceValidator, WorkspaceBackupManager
-from ocrd_utils import getLogger, initLogging, pushd_popd, EXT_TO_MIME
+from ocrd_utils import getLogger, initLogging, pushd_popd, EXT_TO_MIME, safe_filename
 from ocrd.decorators import mets_find_options
 from . import command_with_replaced_help
 
@@ -27,28 +27,10 @@ class WorkspaceCtx():
 
     def __init__(self, directory, mets_url, mets_basename, automatic_backup):
         self.log = getLogger('ocrd.cli.workspace')
-        if mets_basename and mets_url:
-            raise ValueError("Use either --mets or --mets-basename, not both")
-        if mets_basename and not mets_url:
-            self.log.warning(DeprecationWarning("--mets-basename is deprecated. Use --mets/--directory instead"))
-        mets_basename = mets_basename if mets_basename else 'mets.xml'
-        if directory and mets_url:
-            directory = abspath(directory)
-            if not abspath(mets_url).startswith(directory):
-                raise ValueError("--mets has a directory part inconsistent with --directory")
-        elif not directory and mets_url:
-            if mets_url.startswith('http') or mets_url.startswith('https:'):
-                raise ValueError("--mets is an http(s) URL but no --directory was given")
-            directory = dirname(abspath(mets_url)) or getcwd()
-        elif directory and not mets_url:
-            directory = abspath(directory)
-            mets_url = join(directory, mets_basename)
-        else:
-            directory = getcwd()
-            mets_url = join(directory, mets_basename)
-        self.directory = directory
         self.resolver = Resolver()
-        self.mets_url = mets_url
+        if mets_basename:
+            self.log.warning(DeprecationWarning('--mets-basename is deprecated. Use --mets/--directory instead.'))
+        self.directory, self.mets_url, self.mets_basename = self.resolver.resolve_mets_arguments(directory, mets_url, mets_basename)
         self.automatic_backup = automatic_backup
 
 pass_workspace = click.make_pass_decorator(WorkspaceCtx)
@@ -137,8 +119,8 @@ def workspace_clone(ctx, clobber_mets, download, mets_url, workspace_dir):
 
     workspace = ctx.resolver.workspace_from_url(
         mets_url,
-        dst_dir=os.path.abspath(ctx.directory),
-        mets_basename=basename(ctx.mets_url),
+        dst_dir=ctx.directory,
+        mets_basename=ctx.mets_basename,
         clobber_mets=clobber_mets,
         download=download,
     )
@@ -165,8 +147,8 @@ def workspace_init(ctx, clobber_mets, directory):
         LOG.warning(DeprecationWarning("Use 'ocrd workspace --directory DIR init' instead of argument 'DIRECTORY' ('%s')" % directory))
         ctx.directory = directory
     workspace = ctx.resolver.workspace_from_nothing(
-        directory=os.path.abspath(ctx.directory),
-        mets_basename=basename(ctx.mets_url),
+        directory=ctx.directory,
+        mets_basename=ctx.mets_basename,
         clobber_mets=clobber_mets
     )
     workspace.save_mets()
@@ -191,7 +173,7 @@ def workspace_add_file(ctx, file_grp, file_id, mimetype, page_id, ignore, check_
     Add a file or http(s) URL FNAME to METS in a workspace.
     If FNAME is not an http(s) URL and is not a workspace-local existing file, try to copy to workspace.
     """
-    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url), automatic_backup=ctx.automatic_backup)
+    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename, automatic_backup=ctx.automatic_backup)
 
     log = getLogger('ocrd.cli.workspace.add')
     if not mimetype:
@@ -229,7 +211,7 @@ def workspace_add_file(ctx, file_grp, file_id, mimetype, page_id, ignore, check_
     workspace.save_mets()
 
 # ----------------------------------------------------------------------
-# ocrd workspace add-bulk
+# ocrd workspace bulk-add
 # ----------------------------------------------------------------------
 
 # pylint: disable=broad-except
@@ -237,38 +219,58 @@ def workspace_add_file(ctx, file_grp, file_id, mimetype, page_id, ignore, check_
 @click.option('-r', '--regex', help="Regular expression matching the FILE_GLOB filesystem paths to define named captures usable in the other parameters", required=True)
 @click.option('-m', '--mimetype', help="Media type of the file. If not provided, guess from filename", required=False)
 @click.option('-g', '--page-id', help="physical page ID of the file", required=False)
-@click.option('-i', '--file-id', help="ID of the file", required=True)
-@click.option('-u', '--url', help="local filesystem path in the workspace directory (copied from source file if different)", required=True)
+@click.option('-i', '--file-id', help="ID of the file", required=False)
+@click.option('-u', '--url', help="local filesystem path in the workspace directory (copied from source file if different)", required=False)
 @click.option('-G', '--file-grp', help="File group USE of the file", required=True)
 @click.option('-n', '--dry-run', help="Don't actually do anything to the METS or filesystem, just preview", default=False, is_flag=True)
+@click.option('-S', '--source-path', 'src_path_option', help="File path to copy from (if different from FILE_GLOB values)", required=False)
 @click.option('-I', '--ignore', help="Disable checking for existing file entries (faster)", default=False, is_flag=True)
 @click.option('-f', '--force', help="Replace existing file entries with the same ID (no effect when --ignore is set, too)", default=False, is_flag=True)
 @click.option('-s', '--skip', help="Skip files not matching --regex (instead of failing)", default=False, is_flag=True)
 @click.argument('file_glob', nargs=-1, required=True)
 @pass_workspace
-def workspace_cli_bulk_add(ctx, regex, mimetype, page_id, file_id, url, file_grp, dry_run, file_glob, ignore, force, skip):
-    r"""
+def workspace_cli_bulk_add(ctx, regex, mimetype, page_id, file_id, url, file_grp, dry_run, file_glob, src_path_option, ignore, force, skip):
+    """
     Add files in bulk to an OCR-D workspace.
 
-    FILE_GLOB can either be a shell glob expression or a list of files.
+    FILE_GLOB can either be a shell glob expression to match file names,
+    or a list of expressions or '-', in which case expressions are read from STDIN.
 
-    --regex is applied to the absolute path of every file in FILE_GLOB and can
-    define named groups that can be used in --page-id, --file-id, --mimetype, --url and
-    --file-grp by referencing the named group 'grp' in the regex as '{{ grp }}'.
+    After globbing, --regex is matched against each expression resulting from FILE_GLOB, and can
+    define named groups reusable in the --page-id, --file-id, --mimetype, --url, --source-path and
+    --file-grp options, e.g. by referencing the group name 'grp' from the regex as '{{ grp }}'.
+
+    If the FILE_GLOB expressions do not denote the file names themselves
+    (but arbitrary strings for --regex matching), then use --source-path to set
+    the actual file paths to use. (This could involve fixed strings or group references.)
 
     \b
-    Example:
+    Examples:
         ocrd workspace bulk-add \\
-                --regex '^.*/(?P<fileGrp>[^/]+)/page_(?P<pageid>.*)\.(?P<ext>[^\.]*)$' \\
+                --regex '(?P<fileGrp>[^/]+)/page_(?P<pageid>.*)\.[^.]+' \\
+                --page-id 'PHYS_{{ pageid }}' \\
+                --file-grp "{{ fileGrp }}" \\
+                path/to/files/*/*.*
+        \b
+        echo "path/to/src/file.xml SEG/page_p0001.xml" \\
+        | ocrd workspace bulk-add \\
+                --regex '(?P<src>.*?) (?P<fileGrp>.+?)/page_(?P<pageid>.*)\.(?P<ext>[^\.]*)' \\
                 --file-id 'FILE_{{ fileGrp }}_{{ pageid }}' \\
                 --page-id 'PHYS_{{ pageid }}' \\
                 --file-grp "{{ fileGrp }}" \\
                 --url '{{ fileGrp }}/FILE_{{ pageid }}.{{ ext }}' \\
-                path/to/files/*/*.*
+                -
 
+        \b
+        { echo PHYS_0001 BIN FILE_0001_BIN.IMG-wolf BIN/FILE_0001_BIN.IMG-wolf.png; \\
+          echo PHYS_0001 BIN FILE_0001_BIN BIN/FILE_0001_BIN.xml; \\
+          echo PHYS_0002 BIN FILE_0002_BIN.IMG-wolf BIN/FILE_0002_BIN.IMG-wolf.png; \\
+          echo PHYS_0002 BIN FILE_0002_BIN BIN/FILE_0002_BIN.xml; \\
+        } | ocrd workspace bulk-add -r '(?P<pageid>.*) (?P<filegrp>.*) (?P<fileid>.*) (?P<url>.*)' \\
+          -G '{{ filegrp }}' -g '{{ pageid }}' -i '{{ fileid }}' -S '{{ url }}' -
     """
     log = getLogger('ocrd.cli.workspace.bulk-add') # pylint: disable=redefined-outer-name
-    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url), automatic_backup=ctx.automatic_backup)
+    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename, automatic_backup=ctx.automatic_backup)
 
     try:
         pat = re.compile(regex)
@@ -277,8 +279,16 @@ def workspace_cli_bulk_add(ctx, regex, mimetype, page_id, file_id, url, file_grp
         sys.exit(1)
 
     file_paths = []
-    for fglob in file_glob:
-        file_paths += [Path(x).resolve() for x in glob(fglob)]
+    from_stdin = file_glob == ('-',)
+    if from_stdin:
+        file_paths += [Path(x.strip('\n')) for x in sys.stdin.readlines()]
+    else:
+        for fglob in file_glob:
+            expanded = glob(fglob)
+            if not expanded:
+                file_paths += [Path(fglob)]
+            else:
+                file_paths += [Path(x) for x in expanded]
 
     for i, file_path in enumerate(file_paths):
         log.info("[%4d/%d] %s" % (i, len(file_paths), file_path))
@@ -288,10 +298,14 @@ def workspace_cli_bulk_add(ctx, regex, mimetype, page_id, file_id, url, file_grp
         if not m:
             if skip:
                 continue
-            log.error("File not matched by regex: '%s'" % file_path)
+            log.error("File '%s' not matched by regex: '%s'" % (file_path, regex))
             sys.exit(1)
         group_dict = m.groupdict()
 
+        # derive --file-id from filename if not --file-id not explicitly set
+        if not file_id:
+            file_id = safe_filename(str(file_path))
+
         # set up file info
         file_dict = {'url': url, 'mimetype': mimetype, 'ID': file_id, 'pageId': page_id, 'fileGrp': file_grp}
 
@@ -302,20 +316,39 @@ def workspace_cli_bulk_add(ctx, regex, mimetype, page_id, file_id, url, file_grp
             except KeyError:
                 log.error("Cannot guess mimetype from extension '%s' for '%s'. Set --mimetype explicitly" % (file_path.suffix, file_path))
 
+        # Flag to track whether 'url' should be 'src'
+        url_is_src = False
+
         # expand templates
         for param_name in file_dict:
+            if not file_dict[param_name]:
+                if param_name == 'url':
+                    url_is_src = True
+                    continue
+                raise ValueError(f"OcrdFile attribute '{param_name}' unset ({file_dict})")
             for group_name in group_dict:
                 file_dict[param_name] = file_dict[param_name].replace('{{ %s }}' % group_name, group_dict[group_name])
 
-        # copy files
-        if file_dict['url']:
-            urlpath = Path(workspace.directory, file_dict['url'])
-            if not urlpath.exists():
-                log.info("cp '%s' '%s'", file_path, urlpath)
+        # Where to copy from
+        if src_path_option:
+            src_path = src_path_option
+            for group_name in group_dict:
+                src_path = src_path.replace('{{ %s }}' % group_name, group_dict[group_name])
+            srcpath = Path(src_path)
+        else:
+            srcpath = file_path
+
+        # copy files if src != url
+        if url_is_src:
+            file_dict['url'] = str(srcpath)
+        else:
+            destpath = Path(workspace.directory, file_dict['url'])
+            if srcpath != destpath and not destpath.exists():
+                log.info("cp '%s' '%s'", srcpath, destpath)
                 if not dry_run:
-                    if not urlpath.parent.is_dir():
-                        urlpath.parent.mkdir()
-                    urlpath.write_bytes(file_path.read_bytes())
+                    if not destpath.parent.is_dir():
+                        destpath.parent.mkdir()
+                    destpath.write_bytes(srcpath.read_bytes())
 
         # Add to workspace (or not)
         fileGrp = file_dict.pop('fileGrp')
@@ -359,7 +392,7 @@ def workspace_find(ctx, file_grp, mimetype, page_id, file_id, output_field, down
     """
     modified_mets = False
     ret = list()
-    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url))
+    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename)
     for f in workspace.mets.find_files(
             ID=file_id,
             fileGrp=file_grp,
@@ -400,7 +433,7 @@ def workspace_remove_file(ctx, id, force, keep_file):  # pylint: disable=redefin
     (If any ``ID`` starts with ``//``, then its remainder
      will be interpreted as a regular expression.)
     """
-    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url), automatic_backup=ctx.automatic_backup)
+    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename, automatic_backup=ctx.automatic_backup)
     for i in id:
         workspace.remove_file(i, force=force, keep_file=keep_file)
     workspace.save_mets()
@@ -418,7 +451,7 @@ def rename_group(ctx, old, new):
     """
     Rename fileGrp (USE attribute ``NEW`` to ``OLD``).
     """
-    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url))
+    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename)
     workspace.rename_file_group(old, new)
     workspace.save_mets()
 
@@ -439,7 +472,7 @@ def remove_group(ctx, group, recursive, force, keep_files):
     (If any ``GROUP`` starts with ``//``, then its remainder
      will be interpreted as a regular expression.)
     """
-    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url))
+    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename)
     for g in group:
         workspace.remove_file_group(g, recursive=recursive, force=force, keep_files=keep_files)
     workspace.save_mets()
@@ -461,7 +494,7 @@ def prune_files(ctx, file_grp, mimetype, page_id, file_id):
     (If any ``FILTER`` starts with ``//``, then its remainder
      will be interpreted as a regular expression.)
     """
-    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url), automatic_backup=ctx.automatic_backup)
+    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename, automatic_backup=ctx.automatic_backup)
     with pushd_popd(workspace.directory):
         for f in workspace.mets.find_files(
             ID=file_id,
@@ -487,7 +520,7 @@ def list_groups(ctx):
     """
     List fileGrp USE attributes
     """
-    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url))
+    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename)
     print("\n".join(workspace.mets.file_groups))
 
 # ----------------------------------------------------------------------
@@ -500,7 +533,7 @@ def list_pages(ctx):
     """
     List physical page IDs
     """
-    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url))
+    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename)
     print("\n".join(workspace.mets.physical_pages))
 
 # ----------------------------------------------------------------------
@@ -513,7 +546,7 @@ def get_id(ctx):
     """
     Get METS id if any
     """
-    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url))
+    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename)
     ID = workspace.mets.unique_identifier
     if ID:
         print(ID)
@@ -533,7 +566,7 @@ def set_id(ctx, id):   # pylint: disable=redefined-builtin
 
     Otherwise will create a new <mods:identifier type="purl">{{ ID }}</mods:identifier>.
     """
-    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url), automatic_backup=ctx.automatic_backup)
+    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename, automatic_backup=ctx.automatic_backup)
     workspace.mets.unique_identifier = id
     workspace.save_mets()
 
@@ -558,7 +591,7 @@ def merge(ctx, copy_files, filegrp_mapping, file_grp, file_id, page_id, mimetype
     mets_path = Path(mets_path)
     if filegrp_mapping:
         filegrp_mapping = loads(filegrp_mapping)
-    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url), automatic_backup=ctx.automatic_backup)
+    workspace = Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename, automatic_backup=ctx.automatic_backup)
     other_workspace = Workspace(ctx.resolver, directory=str(mets_path.parent), mets_basename=str(mets_path.name))
     workspace.merge(
         other_workspace,
@@ -588,7 +621,7 @@ def workspace_backup_add(ctx):
     """
     Create a new backup
     """
-    backup_manager = WorkspaceBackupManager(Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url), automatic_backup=ctx.automatic_backup))
+    backup_manager = WorkspaceBackupManager(Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename, automatic_backup=ctx.automatic_backup))
     backup_manager.add()
 
 @workspace_backup_cli.command('list')
@@ -597,7 +630,7 @@ def workspace_backup_list(ctx):
     """
     List backups
     """
-    backup_manager = WorkspaceBackupManager(Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url), automatic_backup=ctx.automatic_backup))
+    backup_manager = WorkspaceBackupManager(Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename, automatic_backup=ctx.automatic_backup))
     for b in backup_manager.list():
         print(b)
 
@@ -609,7 +642,7 @@ def workspace_backup_restore(ctx, choose_first, bak):
     """
     Restore backup BAK
     """
-    backup_manager = WorkspaceBackupManager(Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url), automatic_backup=ctx.automatic_backup))
+    backup_manager = WorkspaceBackupManager(Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename, automatic_backup=ctx.automatic_backup))
     backup_manager.restore(bak, choose_first)
 
 @workspace_backup_cli.command('undo')
@@ -618,5 +651,5 @@ def workspace_backup_undo(ctx):
     """
     Restore the last backup
     """
-    backup_manager = WorkspaceBackupManager(Workspace(ctx.resolver, directory=ctx.directory, mets_basename=basename(ctx.mets_url), automatic_backup=ctx.automatic_backup))
+    backup_manager = WorkspaceBackupManager(Workspace(ctx.resolver, directory=ctx.directory, mets_basename=ctx.mets_basename, automatic_backup=ctx.automatic_backup))
     backup_manager.undo()
